@@ -103,23 +103,19 @@ class B2Session(object):
         :param str application_key_id: :term:`application key ID`
         :param str application_key: user's :term:`application key`
         """
-        # Clean up any previous account info if it was for a different account.
-        try:
-            old_account_id = self.account_info.get_account_id()
-            old_realm = self.account_info.get_realm()
-            if application_key_id != old_account_id or realm != old_realm:
-                self.cache.clear()
-        except MissingAccountData:
-            self.cache.clear()
-
         # Authorize
         realm_url = self.account_info.REALM_URLS.get(realm, realm)
         response = self.raw_api.authorize_account(realm_url, application_key_id, application_key)
+        account_id = response['accountId']
         allowed = response['allowed']
 
+        # Clear the cache if new account has been used
+        if not self.account_info.is_same_account(account_id, realm):
+            self.cache.clear()
+
         # Store the auth data
         self.account_info.set_auth_data(
-            account_id=response['accountId'],
+            account_id=account_id,
             auth_token=response['authorizationToken'],
             api_url=response['apiUrl'],
             download_url=response['downloadUrl'],
@@ -177,9 +173,15 @@ class B2Session(object):
     def delete_file_version(self, file_id, file_name):
         return self._wrap_default_token(self.raw_api.delete_file_version, file_id, file_name)
 
-    def download_file_from_url(self, url, range_=None):
+    def download_file_from_url(
+        self, url, range_=None, encryption: Optional[EncryptionSetting] = None
+    ):
         return self._wrap_token(
-            self.raw_api.download_file_from_url, TokenType.API_TOKEN_ONLY, url, range_=range_
+            self.raw_api.download_file_from_url,
+            TokenType.API_TOKEN_ONLY,
+            url,
+            range_=range_,
+            encryption=encryption,
         )
 
     def finish_large_file(self, file_id, part_sha1_array):
@@ -376,6 +378,7 @@ class B2Session(object):
         file_info=None,
         destination_bucket_id=None,
         destination_server_side_encryption: Optional[EncryptionSetting] = None,
+        source_server_side_encryption: Optional[EncryptionSetting] = None,
     ):
         return self._wrap_default_token(
             self.raw_api.copy_file,
@@ -387,6 +390,7 @@ class B2Session(object):
             file_info=file_info,
             destination_bucket_id=destination_bucket_id,
             destination_server_side_encryption=destination_server_side_encryption,
+            source_server_side_encryption=source_server_side_encryption,
         )
 
     def copy_part(
@@ -396,6 +400,7 @@ class B2Session(object):
         part_number,
         bytes_range=None,
         destination_server_side_encryption: Optional[EncryptionSetting] = None,
+        source_server_side_encryption: Optional[EncryptionSetting] = None,
     ):
         return self._wrap_default_token(
             self.raw_api.copy_part,
@@ -404,6 +409,7 @@ class B2Session(object):
             part_number,
             bytes_range=bytes_range,
             destination_server_side_encryption=destination_server_side_encryption,
+            source_server_side_encryption=source_server_side_encryption,
         )
 
     def _wrap_default_token(self, raw_api_method, *args, **kwargs):
