@@ -14,7 +14,7 @@ from enum import Enum, unique
 import logging
 
 from ..exception import DestFileNewer
-from ..encryption.provider import AbstractEncryptionSettingsProvider, SERVER_DEFAULT_ENCRYPTION_SETTINGS_PROVIDER
+from .encryption_provider import AbstractSyncEncryptionSettingsProvider, SERVER_DEFAULT_SYNC_ENCRYPTION_SETTINGS_PROVIDER
 from .action import LocalDeleteAction, B2CopyAction, B2DeleteAction, B2DownloadAction, B2HideAction, B2UploadAction
 from .exception import InvalidArgument
 
@@ -58,7 +58,7 @@ class AbstractFileSyncPolicy(metaclass=ABCMeta):
         compare_threshold,
         compare_version_mode=CompareVersionMode.MODTIME,
         encryption_settings_provider:
-        AbstractEncryptionSettingsProvider = SERVER_DEFAULT_ENCRYPTION_SETTINGS_PROVIDER,
+        AbstractSyncEncryptionSettingsProvider = SERVER_DEFAULT_SYNC_ENCRYPTION_SETTINGS_PROVIDER,
     ):
         """
         :param b2sdk.v1.File source_file: source file object
@@ -70,7 +70,7 @@ class AbstractFileSyncPolicy(metaclass=ABCMeta):
         :param b2sdk.v1.NEWER_FILE_MODES newer_file_mode: setting which determines handling for destination files newer than on the source
         :param int compare_threshold: when comparing with size or time for sync
         :param b2sdk.v1.COMPARE_VERSION_MODES compare_version_mode: how to compare source and destination files
-        :param b2sdk.v1.AbstractEncryptionSettingsProvider encryption_settings_provider: encryption setting provider
+        :param b2sdk.v1.AbstractSyncEncryptionSettingsProvider encryption_settings_provider: encryption setting provider
         """
         self._source_file = source_file
         self._source_folder = source_folder
@@ -225,12 +225,9 @@ class DownPolicy(AbstractFileSyncPolicy):
 
     def _make_transfer_action(self):
         return B2DownloadAction(
-            self._source_file.name,
+            self._source_file,
             self._source_folder.make_full_path(self._source_file.name),
-            self._source_file.latest_version().id_,
             self._dest_folder.make_full_path(self._source_file.name),
-            self._get_source_mod_time(),
-            self._source_file.latest_version().size,
             self._encryption_settings_provider,
         )
 
@@ -320,13 +317,13 @@ class CopyPolicy(AbstractFileSyncPolicy):
     SOURCE_PREFIX = 'b2://'
 
     def _make_transfer_action(self):
+
         return B2CopyAction(
-            self._source_file.name,
             self._source_folder.make_full_path(self._source_file.name),
-            self._source_file.latest_version().id_,
+            self._source_file,
             self._dest_folder.make_full_path(self._source_file.name),
-            self._get_source_mod_time(),
-            self._source_file.latest_version().size,
+            self._source_folder.bucket,
+            self._dest_folder.bucket,
             self._encryption_settings_provider,
         )
 
