@@ -26,6 +26,7 @@ from .deps_exception import (
     InvalidUploadSource,
     MaxRetriesExceeded,
     UnsatisfiableRange,
+    SSECKeyError,
 )
 from .deps import B2Api
 from .deps import LargeFileUploadState
@@ -34,11 +35,40 @@ from .deps import FileVersionInfo
 from .deps import MetadataDirectiveMode
 from .deps import Part
 from .deps import AbstractProgressListener
-from .deps import StubAccountInfo, RawSimulator, BucketSimulator, FakeResponse
+from .deps import StubAccountInfo, RawSimulator, BucketSimulator, FakeResponse, FileSimulator
 from .deps import ParallelDownloader
 from .deps import SimpleDownloader
 from .deps import UploadSourceBytes
 from .deps import hex_sha1_of_bytes, TempDir
+from .deps import EncryptionAlgorithm, EncryptionSetting, EncryptionSettingFactory, EncryptionMode, \
+    EncryptionKey, SSE_NONE, SSE_B2_AES
+from .deps import CopySource, UploadSourceLocalFile, WriteIntent
+
+SSE_C_AES = EncryptionSetting(
+    mode=EncryptionMode.SSE_C,
+    algorithm=EncryptionAlgorithm.AES256,
+    key=EncryptionKey(secret=b'some_key', key_id='some-id'),
+)
+SSE_C_AES_NO_SECRET = EncryptionSetting(
+    mode=EncryptionMode.SSE_C,
+    algorithm=EncryptionAlgorithm.AES256,
+    key=EncryptionKey(secret=None, key_id='some-id'),
+)
+SSE_C_AES_2 = EncryptionSetting(
+    mode=EncryptionMode.SSE_C,
+    algorithm=EncryptionAlgorithm.AES256,
+    key=EncryptionKey(secret=b'some_other_key', key_id='some-id-2'),
+)
+SSE_C_AES_2_NO_SECRET = EncryptionSetting(
+    mode=EncryptionMode.SSE_C,
+    algorithm=EncryptionAlgorithm.AES256,
+    key=EncryptionKey(secret=None, key_id='some-id-2'),
+)
+SSE_C_AES_FROM_SERVER = EncryptionSetting(
+    mode=EncryptionMode.SSE_C,
+    algorithm=EncryptionAlgorithm.AES256,
+    key=EncryptionKey(key_id=None, secret=None),
+)
 
 
 def write_file(path, data):
@@ -143,6 +173,22 @@ class TestCaseWithBucket(TestBase):
         ]
         self.assertEqual(expected, actual)
 
+    def _make_data(self, approximate_length):
+        """
+        Generate a sequence of bytes to use in testing an upload.
+        Don't repeat a short pattern, so we're sure that the different
+        parts of a large file are actually different.
+
+        Returns bytes.
+        """
+        fragments = []
+        so_far = 0
+        while so_far < approximate_length:
+            fragment = ('%d:' % so_far).encode('utf-8')
+            so_far += len(fragment)
+            fragments.append(fragment)
+        return b''.join(fragments)
+
 
 class TestReauthorization(TestCaseWithBucket):
     def testCreateBucket(self):
@@ -227,6 +273,30 @@ class TestListUnfinished(TestCaseWithBucket):
         return self.bucket.start_large_file(file_name, 'text/plain', {})
 
 
+class TestGetFileInfo(TestCaseWithBucket):
+    def test_version_by_name(self):
+        data = b'hello world'
+        a_id = self.bucket.upload_bytes(data, 'a').id_
+
+        info = self.bucket.get_file_info_by_name('a')
+
+        self.assertIsInstance(info, FileVersionInfo)
+        expected = (a_id, 'a', 11, None, 'b2/x-auto')
+        actual = (info.id_, info.file_name, info.size, info.action, info.content_type)
+        self.assertEqual(expected, actual)
+
+    def test_version_by_id(self):
+        data = b'hello world'
+        b_id = self.bucket.upload_bytes(data, 'b').id_
+
+        info = self.bucket.get_file_info_by_id(b_id)
+
+        self.assertIsInstance(info, FileVersionInfo)
+        expected = (b_id, 'b', 11, 'upload', 'b2/x-auto')
+        actual = (info.id_, info.file_name, info.size, info.action, info.content_type)
+        self.assertEqual(expected, actual)
+
+
 class TestLs(TestCaseWithBucket):
     def test_empty(self):
         self.assertEqual([], list(self.bucket.ls('foo')))
@@ -311,6 +381,42 @@ class TestLs(TestCaseWithBucket):
         self.assertBucketContents(expected, '', show_versions=True)
 
 
+class TestListVersions(TestCaseWithBucket):
+    def test_encryption(self):
+        data = b'hello world'
+        a = self.bucket.upload_bytes(data, 'a')
+        a_id = a.id_
+        self.assertEqual(a.server_side_encryption, SSE_NONE)
+        b = self.bucket.upload_bytes(data, 'b', encryption=SSE_B2_AES)
+        self.assertEqual(b.server_side_encryption, SSE_B2_AES)
+        b_id = b.id_
+        # c_id = self.bucket.upload_bytes(data, 'c', encryption=SSE_NONE).id_  # TODO
+        self.bucket.copy(a_id, 'd', destination_encryption=SSE_B2_AES)
+        self.bucket.copy(
+            b_id, 'e', destination_encryption=SSE_C_AES, file_info={}, content_type='text/plain'
+        )
+
+        actual = [info for info in self.bucket.list_file_versions('a')['files']][0]
+        actual = EncryptionSettingFactory.from_file_version_dict(actual)
+        self.assertEqual(SSE_NONE, actual)  # bucket default
+
+        actual = [info for info in self.bucket.list_file_versions('b')['files']][0]
+        actual = EncryptionSettingFactory.from_file_version_dict(actual)
+        self.assertEqual(SSE_B2_AES, actual)  # explicitly requested sse-b2
+
+        # actual = [info for info in self.bucket.list_file_versions('c')][0]
+        # actual = EncryptionSettingFactory.from_file_version_dict(actual)
+        # self.assertEqual(SSE_NONE, actual)  # explicitly requested none
+
+        actual = [info for info in self.bucket.list_file_versions('d')['files']][0]
+        actual = EncryptionSettingFactory.from_file_version_dict(actual)
+        self.assertEqual(SSE_B2_AES, actual)  # explicitly requested sse-b2
+
+        actual = [info for info in self.bucket.list_file_versions('e')['files']][0]
+        actual = EncryptionSettingFactory.from_file_version_dict(actual)
+        self.assertEqual(SSE_C_AES_NO_SECRET, actual)  # explicitly requested sse-c
+
+
 class TestCopyFile(TestCaseWithBucket):
     def test_copy_without_optional_params(self):
         file_id = self._make_file()
@@ -410,6 +516,129 @@ class TestCopyFile(TestCaseWithBucket):
         expected = [('hello_new.txt', 11, 'copy', None)]
         self.assertBucketContents(expected, '', show_versions=True)
 
+    def test_copy_encryption(self):
+        data = b'hello_world'
+        a = self.bucket.upload_bytes(data, 'a')
+        a_id = a.id_
+        self.assertEqual(a.server_side_encryption, SSE_NONE)
+
+        b = self.bucket.upload_bytes(data, 'b', encryption=SSE_B2_AES)
+        self.assertEqual(b.server_side_encryption, SSE_B2_AES)
+        b_id = b.id_
+
+        c = self.bucket.upload_bytes(data, 'c', encryption=SSE_C_AES)
+        self.assertEqual(c.server_side_encryption, SSE_C_AES_NO_SECRET)
+        c_id = c.id_
+
+        for length in [None, len(data)]:
+            for kwargs, expected_encryption in [
+                (dict(file_id=a_id, destination_encryption=SSE_B2_AES), SSE_B2_AES),
+                (
+                    dict(
+                        file_id=a_id,
+                        destination_encryption=SSE_C_AES,
+                        file_info={'new': 'value'},
+                        content_type='text/plain'
+                    ), SSE_C_AES_NO_SECRET
+                ),
+                (
+                    dict(
+                        file_id=a_id,
+                        destination_encryption=SSE_C_AES,
+                        source_file_info={'old': 'value'},
+                        source_content_type='text/plain'
+                    ), SSE_C_AES_NO_SECRET
+                ),
+                (dict(file_id=b_id), SSE_NONE),
+                (dict(file_id=b_id, source_encryption=SSE_B2_AES), SSE_NONE),
+                (
+                    dict(
+                        file_id=b_id,
+                        source_encryption=SSE_B2_AES,
+                        destination_encryption=SSE_B2_AES
+                    ), SSE_B2_AES
+                ),
+                (
+                    dict(
+                        file_id=b_id,
+                        source_encryption=SSE_B2_AES,
+                        destination_encryption=SSE_C_AES,
+                        file_info={'new': 'value'},
+                        content_type='text/plain'
+                    ), SSE_C_AES_NO_SECRET
+                ),
+                (
+                    dict(
+                        file_id=b_id,
+                        source_encryption=SSE_B2_AES,
+                        destination_encryption=SSE_C_AES,
+                        source_file_info={'old': 'value'},
+                        source_content_type='text/plain'
+                    ), SSE_C_AES_NO_SECRET
+                ),
+                (
+                    dict(
+                        file_id=c_id,
+                        source_encryption=SSE_C_AES,
+                        file_info={'new': 'value'},
+                        content_type='text/plain'
+                    ), SSE_NONE
+                ),
+                (
+                    dict(
+                        file_id=c_id,
+                        source_encryption=SSE_C_AES,
+                        source_file_info={'old': 'value'},
+                        source_content_type='text/plain'
+                    ), SSE_NONE
+                ),
+                (
+                    dict(
+                        file_id=c_id, source_encryption=SSE_C_AES, destination_encryption=SSE_C_AES
+                    ), SSE_C_AES_NO_SECRET
+                ),
+                (
+                    dict(
+                        file_id=c_id,
+                        source_encryption=SSE_C_AES,
+                        destination_encryption=SSE_B2_AES,
+                        source_file_info={'old': 'value'},
+                        source_content_type='text/plain'
+                    ), SSE_B2_AES
+                ),
+                (
+                    dict(
+                        file_id=c_id,
+                        source_encryption=SSE_C_AES,
+                        destination_encryption=SSE_B2_AES,
+                        file_info={'new': 'value'},
+                        content_type='text/plain'
+                    ), SSE_B2_AES
+                ),
+                (
+                    dict(
+                        file_id=c_id,
+                        source_encryption=SSE_C_AES,
+                        destination_encryption=SSE_C_AES_2,
+                        source_file_info={'old': 'value'},
+                        source_content_type='text/plain'
+                    ), SSE_C_AES_2_NO_SECRET
+                ),
+                (
+                    dict(
+                        file_id=c_id,
+                        source_encryption=SSE_C_AES,
+                        destination_encryption=SSE_C_AES_2,
+                        file_info={'new': 'value'},
+                        content_type='text/plain'
+                    ), SSE_C_AES_2_NO_SECRET
+                ),
+            ]:
+                with self.subTest(kwargs=kwargs, length=length, data=data):
+                    file_info = self.bucket.copy(**kwargs, new_file_name='new_file', length=length)
+                    self.assertTrue(isinstance(file_info, FileVersionInfo))
+                    self.assertEqual(file_info.server_side_encryption, expected_encryption)
+
     def _make_file(self, bucket=None):
         data = b'hello world'
         actual_bucket = bucket or self.bucket
@@ -423,6 +652,38 @@ class TestUpload(TestCaseWithBucket):
         self.assertTrue(isinstance(file_info, FileVersionInfo))
         self._check_file_contents('file1', data)
 
+    def test_upload_bytes_sse_b2(self):
+        data = b'hello world'
+        file_info = self.bucket.upload_bytes(data, 'file1', encryption=SSE_B2_AES)
+        self.assertTrue(isinstance(file_info, FileVersionInfo))
+        self.assertEqual(file_info.server_side_encryption, SSE_B2_AES)
+
+    def test_upload_bytes_sse_c(self):
+        data = b'hello world'
+        file_info = self.bucket.upload_bytes(data, 'file1', encryption=SSE_C_AES)
+        self.assertTrue(isinstance(file_info, FileVersionInfo))
+        self.assertEqual(SSE_C_AES_NO_SECRET, file_info.server_side_encryption)
+
+    def test_upload_local_file_sse_b2(self):
+        with TempDir() as d:
+            path = os.path.join(d, 'file1')
+            data = b'hello world'
+            write_file(path, data)
+            file_info = self.bucket.upload_local_file(path, 'file1', encryption=SSE_B2_AES)
+            self.assertTrue(isinstance(file_info, FileVersionInfo))
+            self.assertEqual(file_info.server_side_encryption, SSE_B2_AES)
+            self._check_file_contents('file1', data)
+
+    def test_upload_local_file_sse_c(self):
+        with TempDir() as d:
+            path = os.path.join(d, 'file1')
+            data = b'hello world'
+            write_file(path, data)
+            file_info = self.bucket.upload_local_file(path, 'file1', encryption=SSE_C_AES)
+            self.assertTrue(isinstance(file_info, FileVersionInfo))
+            self.assertEqual(SSE_C_AES_NO_SECRET, file_info.server_side_encryption)
+            self._check_file_contents('file1', data)
+
     def test_upload_bytes_progress(self):
         data = b'hello world'
         progress_listener = StubProgressListener()
@@ -579,24 +840,91 @@ class TestUpload(TestCaseWithBucket):
 
     def _check_file_contents(self, file_name, expected_contents):
         download = DownloadDestBytes()
-        self.bucket.download_file_by_name(file_name, download)
+        with FileSimulator.dont_check_encryption():
+            self.bucket.download_file_by_name(file_name, download)
         self.assertEqual(expected_contents, download.get_bytes_written())
 
-    def _make_data(self, approximate_length):
-        """
-        Generate a sequence of bytes to use in testing an upload.
-        Don't repeat a short pattern, so we're sure that the different
-        parts of a large file are actually different.
 
-        Returns bytes.
-        """
-        fragments = []
-        so_far = 0
-        while so_far < approximate_length:
-            fragment = ('%d:' % so_far).encode('utf-8')
-            so_far += len(fragment)
-            fragments.append(fragment)
-        return b''.join(fragments)
+class TestConcatenate(TestCaseWithBucket):
+    def _create_remote(self, sources, file_name, encryption=None):
+        return self.bucket.concatenate(sources, file_name=file_name, encryption=encryption)
+
+    def test_create_remote(self):
+        data = b'hello world'
+
+        f1_id = self.bucket.upload_bytes(data, 'f1').id_
+        f2_id = self.bucket.upload_bytes(data, 'f1').id_
+        with TempDir() as d:
+            path = os.path.join(d, 'file')
+            write_file(path, data)
+            created_file = self._create_remote(
+                [
+                    CopySource(f1_id, length=len(data), offset=0),
+                    UploadSourceLocalFile(path),
+                    CopySource(f2_id, length=len(data), offset=0),
+                ],
+                file_name='created_file'
+            )
+            self.assertIsInstance(created_file, FileVersionInfo)
+            actual = (
+                created_file.id_, created_file.file_name, created_file.size,
+                created_file.server_side_encryption
+            )
+            expected = ('9997', 'created_file', 33, SSE_NONE)
+            self.assertEqual(expected, actual)
+
+    def test_create_remote_encryption(self):
+        for data in [b'hello_world', self._make_data(self.simulator.MIN_PART_SIZE * 3)]:
+            f1_id = self.bucket.upload_bytes(data, 'f1', encryption=SSE_C_AES).id_
+            f2_id = self.bucket.upload_bytes(data, 'f1', encryption=SSE_C_AES_2).id_
+            with TempDir() as d:
+                path = os.path.join(d, 'file')
+                write_file(path, data)
+                created_file = self._create_remote(
+                    [
+                        CopySource(f1_id, length=len(data), offset=0, encryption=SSE_C_AES),
+                        UploadSourceLocalFile(path),
+                        CopySource(f2_id, length=len(data), offset=0, encryption=SSE_C_AES_2),
+                    ],
+                    file_name='created_file_%s' % (len(data),),
+                    encryption=SSE_C_AES
+                )
+            self.assertIsInstance(created_file, FileVersionInfo)
+            actual = (
+                created_file.id_, created_file.file_name, created_file.size,
+                created_file.server_side_encryption
+            )
+            expected = (
+                mock.ANY,
+                'created_file_%s' % (len(data),),
+                mock.ANY,  # FIXME: this should be equal to len(data) * 3,
+                # but there is a problem in the simulator/test code somewhere
+                SSE_C_AES_NO_SECRET
+            )
+            self.assertEqual(expected, actual)
+
+
+class TestCreateFile(TestConcatenate):
+    def _create_remote(self, sources, file_name, encryption=None):
+        return self.bucket.create_file(
+            [wi for wi in WriteIntent.wrap_sources_iterator(sources)],
+            file_name=file_name,
+            encryption=encryption
+        )
+
+
+class TestConcatenateStream(TestConcatenate):
+    def _create_remote(self, sources, file_name, encryption=None):
+        return self.bucket.concatenate_stream(sources, file_name=file_name, encryption=encryption)
+
+
+class TestCreateFileStream(TestConcatenate):
+    def _create_remote(self, sources, file_name, encryption=None):
+        return self.bucket.create_file_stream(
+            [wi for wi in WriteIntent.wrap_sources_iterator(sources)],
+            file_name=file_name,
+            encryption=encryption
+        )
 
 
 # Downloads
@@ -608,6 +936,9 @@ class DownloadTests(object):
     def setUp(self):
         super(DownloadTests, self).setUp()
         self.file_info = self.bucket.upload_bytes(self.DATA.encode(), 'file1')
+        self.encrypted_file_info = self.bucket.upload_bytes(
+            self.DATA.encode(), 'enc_file1', encryption=SSE_C_AES
+        )
         self.download_dest = DownloadDestBytes()
         self.progress_listener = StubProgressListener()
 
@@ -720,6 +1051,18 @@ class DownloadTests(object):
             )
             self._check_local_file_contents(path, b'1234567defghij567890')
 
+    def test_download_by_id_no_progress_encryption(self):
+        self.bucket.download_file_by_id(
+            self.encrypted_file_info.id_, self.download_dest, encryption=SSE_C_AES
+        )
+        self._verify(self.DATA, check_progress_listener=False)
+
+    def test_download_by_id_no_progress_wrong_encryption(self):
+        with self.assertRaises(SSECKeyError):
+            self.bucket.download_file_by_id(
+                self.encrypted_file_info.id_, self.download_dest, encryption=SSE_C_AES_2
+            )
+
     def _check_local_file_contents(self, path, expected_contents):
         with open(path, 'rb') as f:
             contents = f.read()
